"""
Send a transaction from console, with no password nor confirmation asked.
To be used for unattended, automated processes.

This file takes optional arguments,

arg1: amount to send
arg2: recipient address
arg3: operation
arg4: OpenField data
arg5: wallet file
arg6: request confirmation for every transaction

args3,4,6 are not prompted if ran without args
"""

import base64
import sys
import time

from Cryptodome.Hash import SHA
from Cryptodome.Signature import PKCS1_v1_5

from bismuthclient import rpcconnections
from bisbasic import essentials, options
from bisbasic.essentials import fee_calculate
from polysign.signerfactory import SignerFactory


def connect():
    if 'regnet' in config.version:
        port = 3030
    elif 'testnet' in config.version:
        port = 2829
    else:
        port = 5658

    return rpcconnections.Connection(('127.0.0.1', int(port)))


if __name__ == '__main__':
    config = options.Get()
    config.read()

    try:
        wallet_file = sys.argv[5]
    except:
        wallet_file = input('Path to wallet: ')

    try:
        request_confirmation = sys.argv[6]
    except:
        request_confirmation = False

    key, public_key_readable, private_key_readable, encrypted, unlocked, public_key_b64encoded, address, keyfile = essentials.keys_load_new(wallet_file)

    if encrypted:
        key, private_key_readable = essentials.keys_unlock(private_key_readable)

    print(f'Number of arguments: {len(sys.argv)} arguments.')
    print(f'Argument list: {"".join(sys.argv)}')
    print(f'Using address: {address}')

    # get balance

    s = connect()
    s._send('balanceget')
    s._send(address)  # change address here to view other people's transactions
    stats_account = s._receive()
    balance = stats_account[0]

    print('Transaction address: %s' % address)
    print('Transaction address balance: %s' % balance)

    try:
        amount_input = sys.argv[1]
    except IndexError:
        amount_input = input('Amount: ')

    try:
        recipient_input = sys.argv[2]
    except IndexError:
        recipient_input = input('Recipient: ')

    if not SignerFactory.address_is_valid(recipient_input):
        print('Wrong address format')
        sys.exit(1)

    try:
        operation_input = sys.argv[3]
    except IndexError:
        operation_input = ''

    try:
        openfield_input = sys.argv[4]
    except IndexError:
        openfield_input = ''

    fee = fee_calculate(openfield_input)
    print('Fee: %s' % fee)

    if request_confirmation:
        confirm = input('Confirm (y/n): ')

        if confirm != 'y':
            print('Transaction cancelled, user confirmation failed')
            exit(1)

    try:
        float(amount_input)
        is_float = 1
    except ValueError:
        is_float = 0
        sys.exit(1)

    timestamp = '%.2f' % (time.time() - 5)  #remote proofing
    # TODO: use transaction object, no dup code for buffer assembling
    transaction = (str(timestamp), str(address), str(recipient_input), '%.8f' % float(amount_input), str(operation_input), str(openfield_input))  # this is signed
    # TODO: use polysign here
    h = SHA.new(str(transaction).encode('utf-8'))
    signer = PKCS1_v1_5.new(key)
    signature = signer.sign(h)
    signature_enc = base64.b64encode(signature)
    txid = signature_enc[:56]

    print(f"Encoded Signature: {signature_enc.decode('utf-8')}")
    print(f"Transaction ID: {txid.decode('utf-8')}")

    verifier = PKCS1_v1_5.new(key)

    if verifier.verify(h, signature):
        if float(amount_input) < 0:
            print('Signature OK, but cannot use negative amounts')

        elif float(amount_input) + float(fee) > float(balance):
            print('Mempool: Sending more than owned')

        else:
            tx_submit = (str(timestamp), str(address), str(recipient_input), '%.8f' % float(amount_input), str(signature_enc.decode('utf-8')), str(public_key_b64encoded.decode('utf-8')), str(operation_input), str(openfield_input))
            while True:
                try:
                    s._send('mpinsert')
                    s._send(tx_submit)
                    reply = s._receive()
                    print('Client: {}'.format(reply))
                    if reply != '*':  # response can be empty due to different timeout setting
                        break
                    else:
                        print('Connection cut, retrying')

                except Exception as e:
                    print(f'A problem occurred: {e}, retrying')
                    s = connect()
    else:
        print('Invalid signature')

    s.close()



#!/usr/bin/env python
# list_files.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (list_files.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com

#------------------------------------------------------------------------------

from __future__ import absolute_import
from six.moves import map
from six.moves import range
from io import StringIO

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

import os
import zlib

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.system import bpio

from bitdust.lib import strng
from bitdust.lib import packetid
from bitdust.lib import misc

from bitdust.main import settings

from bitdust.crypt import encrypted
from bitdust.crypt import key
from bitdust.crypt import my_keys

from bitdust.p2p import commands
from bitdust.p2p import p2p_service
from bitdust.contacts import identitycache

from bitdust.userid import my_id
from bitdust.userid import global_id

#------------------------------------------------------------------------------


def send(customer_idurl, packet_id, format_type, key_id, remote_idurl, query_items=[]):
    if not query_items:
        query_items = ['*']
    key_id = my_keys.latest_key_id(key_id)
    parts = global_id.NormalizeGlobalID(key_id)
    if parts['key_alias'] == 'master' and parts['idurl'] != my_id.getIDURL():
        # lg.warn('incoming ListFiles() request with customer "master" key: %r' % key_id)
        if not my_keys.is_key_registered(key_id) and identitycache.HasKey(parts['idurl']):
            lg.info('customer public key %r to be registered locally for the first time' % key_id)
            known_ident = identitycache.FromCache(parts['idurl'])
            if not my_keys.register_key(key_id, known_ident.getPublicKey()):
                lg.err('failed to register known public key of the customer: %r' % key_id)
    if not my_keys.is_key_registered(key_id):
        lg.warn('not able to return Files() for customer %s, key %s not registered' % (customer_idurl, key_id))
        return p2p_service.SendFailNoRequest(customer_idurl, packet_id, response='key not registered')
    if _Debug:
        lg.out(_DebugLevel, 'list_files.send to %s, customer_idurl=%s, key_id=%s, query_items=%r' % (remote_idurl, customer_idurl, key_id, query_items))
    ownerdir = settings.getCustomerFilesDir(customer_idurl)
    plaintext = ''
    if os.path.isdir(ownerdir):
        try:
            for query_path in query_items:
                plaintext += process_query_item(query_path, parts['key_alias'], ownerdir)
        except:
            lg.exc()
            return p2p_service.SendFailNoRequest(customer_idurl, packet_id, response='list files query processing error')
    else:
        lg.warn('did not found customer folder: %s' % ownerdir)
    if _Debug:
        lg.out(_DebugLevel, '\n%s' % plaintext)
    raw_list_files = PackListFiles(plaintext, format_type)
    block = encrypted.Block(
        CreatorID=my_id.getIDURL(),
        BackupID=key_id,
        Data=raw_list_files,
        SessionKey=key.NewSessionKey(session_key_type=key.SessionKeyType()),
        SessionKeyType=key.SessionKeyType(),
        EncryptKey=key_id,
    )
    encrypted_list_files = block.Serialize()
    newpacket = p2p_service.SendFiles(
        idurl=remote_idurl,
        raw_list_files_info=encrypted_list_files,
        packet_id=packet_id,
        callbacks={
            commands.Ack(): on_acked,
            commands.Fail(): on_failed,
            None: on_timeout,
        },
    )
    return newpacket


def process_query_item(query_path, key_alias, ownerdir):
    ret = ''
    ret += 'Q%s\n' % query_path
    if query_path == '*':
        if key_alias == 'master':
            key_alias_dir = os.path.join(ownerdir, key_alias)
            ret += TreeSummary(key_alias_dir, key_alias)
        for one_key_alias in os.listdir(ownerdir):
            if one_key_alias == 'master':
                continue
            if key_alias and key_alias != 'master' and one_key_alias != key_alias:
                continue
            if not misc.ValidKeyAlias(strng.to_text(one_key_alias)):
                continue
            key_alias_dir = os.path.join(ownerdir, one_key_alias)
            ret += TreeSummary(key_alias_dir, one_key_alias)
        if _Debug:
            lg.args(_DebugLevel, o=ownerdir, q=query_path, k=key_alias, result_bytes=len(ret))
        return ret
    # TODO: more validations to be added
    clean_path = query_path.replace('.', '').replace('~', '').replace(':', '').replace('\\', '/').lstrip('/')
    path_items = clean_path.split('/')
    path_items.insert(0, ownerdir)
    local_path = os.path.join(*path_items)
    if not os.path.exists(local_path):
        lg.warn('local file or folder not exist: %r' % local_path)
        return ''
    ret += TreeSummary(local_path, key_alias)
    if _Debug:
        lg.args(_DebugLevel, o=ownerdir, q=query_path, k=key_alias, p=local_path, result_bytes=len(ret))
    return ret


#------------------------------------------------------------------------------


def on_acked(response, info):
    if _Debug:
        lg.out(_DebugLevel, 'list_files.on_acked with %s in %s' % (response, info))


def on_failed(response, error):
    lg.warn('send files %s failed with %s' % (response, error))


def on_timeout(pkt_out):
    lg.warn('send files with %s was timed out' % pkt_out)


#------------------------------------------------------------------------------


def PackListFiles(plaintext, method):
    if method == 'Text':
        return plaintext
    elif method == 'Compressed':
        return zlib.compress(strng.to_bin(plaintext))
    return ''


#------------------------------------------------------------------------------


def TreeSummary(ownerdir, key_alias):
    out = StringIO()
    out.write('K%s\n' % key_alias)

    def cb(result, realpath, subpath, name):
        if not os.access(realpath, os.R_OK):
            return False
        if os.path.isfile(realpath):
            try:
                filesz = os.path.getsize(realpath)
            except:
                filesz = -1
            result.write('F%s %d\n' % (subpath, filesz))
            return False
        if not packetid.IsCanonicalVersion(name):
            found_some_versions = False
            for sub_path in os.listdir(realpath):
                if packetid.IsCanonicalVersion(sub_path):
                    found_some_versions = True
                    break
            if found_some_versions:
                result.write('F%s -1\n' % subpath)
            else:
                result.write('D%s\n' % subpath)
            return True
        maxBlock = -1
        versionSize = {}
        dataBlocks = {}
        parityBlocks = {}
        dataMissing = {}
        parityMissing = {}
        for filename in os.listdir(realpath):
            packetID = subpath + '/' + filename
            pth = os.path.join(realpath, filename)
            if os.path.isdir(pth):
                result.write('D%s\n' % packetID)
                continue
            try:
                filesz = os.path.getsize(pth)
            except:
                filesz = -1
            if not packetid.Valid(packetID):
                result.write('F%s %d\n' % (packetID, filesz))
                continue
            _, pathID, versionName, blockNum, supplierNum, dataORparity = packetid.SplitFull(packetID)
            if None in [pathID, versionName, blockNum, supplierNum, dataORparity]:
                result.write('F%s %d\n' % (packetID, filesz))
                continue
            if dataORparity != 'Data' and dataORparity != 'Parity':
                result.write('F%s %d\n' % (packetID, filesz))
                continue
            if maxBlock < blockNum:
                maxBlock = blockNum
            if supplierNum not in versionSize:
                versionSize[supplierNum] = 0
            if supplierNum not in dataBlocks:
                dataBlocks[supplierNum] = {}
            if supplierNum not in parityBlocks:
                parityBlocks[supplierNum] = {}
            if dataORparity == 'Data':
                dataBlocks[supplierNum][blockNum] = filesz
            elif dataORparity == 'Parity':
                parityBlocks[supplierNum][blockNum] = filesz
        for supplierNum in versionSize.keys():
            dataMissing[supplierNum] = set(range(maxBlock + 1))
            parityMissing[supplierNum] = set(range(maxBlock + 1))
            for blockNum in range(maxBlock + 1):
                if blockNum in list(dataBlocks[supplierNum].keys()):
                    versionSize[supplierNum] += dataBlocks[supplierNum][blockNum]
                    dataMissing[supplierNum].discard(blockNum)
                if blockNum in list(parityBlocks[supplierNum].keys()):
                    versionSize[supplierNum] += parityBlocks[supplierNum][blockNum]
                    parityMissing[supplierNum].discard(blockNum)
        suppliers = set(list(dataBlocks.keys()) + list(parityBlocks.keys()))
        for supplierNum in suppliers:
            versionString = '%s %d 0-%d %d' % (subpath, supplierNum, maxBlock, versionSize[supplierNum])
            if len(dataMissing[supplierNum]) > 0 or len(parityMissing[supplierNum]) > 0:
                versionString += ' missing'
                if len(dataMissing[supplierNum]) > 0:
                    versionString += ' Data:' + (','.join(map(str, dataMissing[supplierNum])))
                if len(parityMissing[supplierNum]) > 0:
                    versionString += ' Parity:' + (','.join(map(str, parityMissing[supplierNum])))
            result.write('V%s\n' % versionString)
        del dataBlocks
        del parityBlocks
        del dataMissing
        del parityMissing
        return False

    bpio.traverse_dir_recursive(lambda realpath, subpath, name: cb(out, realpath, subpath, name), ownerdir)
    src = out.getvalue()
    out.close()
    return src


#------------------------------------------------------------------------------



#locally stored to prevent an even of abandoned repository removal

from Cryptodome.Cipher import AES
from Cryptodome.Hash import SHA256, HMAC
from Cryptodome.Protocol.KDF import PBKDF2
from Cryptodome.Random.random import getrandbits
from Cryptodome.Util import Counter

# see: http://www.daemonology.net/blog/2009-06-11-cryptographic-right-answers.html

EXPANSION_COUNT = (10000, 10000, 100000)
AES_KEY_LEN = 256
SALT_LEN = (128, 256, 256)
HASH = SHA256
PREFIX = b'sc'
HEADER = (PREFIX + b'\x00\x00', PREFIX + b'\x00\x01', PREFIX + b'\x00\x02')
LATEST = 2  # index into SALT_LEN, EXPANSION_COUNT, HEADER

# lengths here are in bits, but pcrypto uses block size in bytes
HALF_BLOCK = AES.block_size*8 // 2
for salt_len in SALT_LEN:
    assert HALF_BLOCK <= salt_len  # we use a subset of the salt as nonce

HEADER_LEN = 4
for header in HEADER:
    assert len(header) == HEADER_LEN


def encrypt(password, data):
    '''
    Encrypt some data.  Input can be bytes or a string (which will be encoded
    using UTF-8).

    @param password: The secret value used as the basis for a key.
    This should be as long as varied as possible.  Try to avoid common words.

    @param data: The data to be encrypted.

    @return: The encrypted data, as bytes.
    '''
    data = _str_to_bytes(data)
    _assert_encrypt_length(data)
    salt = bytes(_random_bytes(SALT_LEN[LATEST] // 8))
    hmac_key, cipher_key = _expand_keys(password, salt, EXPANSION_COUNT[LATEST])
    counter = Counter.new(HALF_BLOCK, prefix=salt[:HALF_BLOCK // 8])
    cipher = AES.new(cipher_key, AES.MODE_CTR, counter=counter)
    encrypted = cipher.encrypt(data)
    hmac = _hmac(hmac_key, HEADER[LATEST] + salt + encrypted)
    return HEADER[LATEST] + salt + encrypted + hmac


def decrypt(password, data):
    '''
    Decrypt some data.  Input must be bytes.

    @param password: The secret value used as the basis for a key.
    This should be as long as varied as possible.  Try to avoid common words.

    @param data: The data to be decrypted, typically as bytes.

    @return: The decrypted data, as bytes.  If the original message was a
    string you can re-create that using `result.decode('utf8')`.
    '''
    _assert_not_unicode(data)
    _assert_header_prefix(data)
    version = _assert_header_version(data)
    _assert_decrypt_length(data, version)
    raw = data[HEADER_LEN:]
    salt = raw[:SALT_LEN[version] // 8]
    hmac_key, cipher_key = _expand_keys(password, salt, EXPANSION_COUNT[version])
    hmac = raw[-HASH.digest_size:]
    hmac2 = _hmac(hmac_key, data[:-HASH.digest_size])
    _assert_hmac(hmac_key, hmac, hmac2)
    counter = Counter.new(HALF_BLOCK, prefix=salt[:HALF_BLOCK // 8])
    cipher = AES.new(cipher_key, AES.MODE_CTR, counter=counter)
    return cipher.decrypt(raw[SALT_LEN[version] // 8:-HASH.digest_size])


class DecryptionException(Exception):
    pass


class EncryptionException(Exception):
    pass


def _assert_not_unicode(data):
    # warn confused users
    u_type = type(b''.decode('utf8'))
    if isinstance(data, u_type):
        raise DecryptionException('Data to decrypt must be bytes; ' + 'you cannot use a string because no string encoding will accept all possible characters.')


def _assert_encrypt_length(data):
    # for AES this is never going to fail
    if len(data) > 2**HALF_BLOCK:
        raise EncryptionException('Message too long.')


def _assert_decrypt_length(data, version):
    if len(data) < HEADER_LEN + SALT_LEN[version] // 8 + HASH.digest_size:
        raise DecryptionException('Missing data.')


def _assert_header_prefix(data):
    if len(data) >= 2 and data[:2] != PREFIX:
        raise DecryptionException('Data passed to decrypt were not generated by simple-crypt (bad header).')


def _assert_header_version(data):
    if len(data) >= HEADER_LEN:
        try:
            return HEADER.index(data[:HEADER_LEN])
        except:
            raise DecryptionException('The data appear to be encrypted with a more recent version of simple-crypt (bad header). ' + 'Please update the library and try again.')
    else:
        raise DecryptionException('Missing header.')


def _assert_hmac(key, hmac, hmac2):
    # https://www.isecpartners.com/news-events/news/2011/february/double-hmac-verification.aspx
    if _hmac(key, hmac) != _hmac(key, hmac2):
        raise DecryptionException('Bad password or corrupt / modified data.')


def _pbkdf2(password, salt, n_bytes, count):
    # the form of the prf below is taken from the code for PBKDF2
    return PBKDF2(password, salt, dkLen=n_bytes, count=count, prf=lambda p, s: HMAC.new(p, s, HASH).digest())


def _expand_keys(password, salt, expansion_count):
    if not salt:
        raise ValueError('Missing salt.')
    if not password:
        raise ValueError('Missing password.')
    key_len = AES_KEY_LEN // 8
    keys = _pbkdf2(_str_to_bytes(password), salt, 2*key_len, expansion_count)
    return keys[:key_len], keys[key_len:]


def _hide(ranbytes):
    # appelbaum recommends obscuring output from random number generators since it can reveal state.
    # we can do this explicitly with a hash, but this is what a PBKDF does anyway, so use one.
    # we don't care about the salt or work factor because there is a large space of values anyway.
    return bytearray(_pbkdf2(bytes(ranbytes), b'', len(ranbytes), 1))


def _random_bytes(n):
    return _hide(bytearray(getrandbits(8) for _ in range(n)))


def _hmac(key, data):
    return HMAC.new(key, data, HASH).digest()


def _str_to_bytes(data):
    u_type = type(b''.decode('utf8'))
    if isinstance(data, u_type):
        return data.encode('utf8')
    return data



#!/usr/bin/env python
# keys_synchronizer.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (keys_synchronizer.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: keys_synchronizer
.. role:: red

BitDust keys_synchronizer() Automat

EVENTS:
    * :red:`backup-ok`
    * :red:`clean-ok`
    * :red:`disconnected`
    * :red:`error`
    * :red:`init`
    * :red:`instant`
    * :red:`restore-ok`
    * :red:`run`
    * :red:`shutdown`
    * :red:`sync`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

import sys

#------------------------------------------------------------------------------

try:
    from twisted.internet import reactor  # @UnresolvedImport
except:
    sys.exit('Error initializing twisted.internet.reactor in keys_synchronizer.py')

#------------------------------------------------------------------------------

from twisted.internet.defer import Deferred, DeferredList
from twisted.python import failure

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.automats import automat

from bitdust.crypt import my_keys

from bitdust.storage import backup_fs
from bitdust.storage import restore_monitor

from bitdust.main import events

from bitdust.raid import eccmap

from bitdust.access import key_ring

#------------------------------------------------------------------------------

_KeysSynchronizer = None

#------------------------------------------------------------------------------


def is_synchronized():
    if not A():
        return False
    return A().state == 'IN_SYNC!'


#------------------------------------------------------------------------------


def A(event=None, *args, **kwargs):
    """
    Access method to interact with `keys_synchronizer()` machine.
    """
    global _KeysSynchronizer
    if event is None:
        return _KeysSynchronizer
    if _KeysSynchronizer is None:
        _KeysSynchronizer = KeysSynchronizer(
            name='keys_synchronizer',
            state='AT_STARTUP',
            debug_level=_DebugLevel,
            log_events=_Debug,
            log_transitions=_Debug,
        )
    if event is not None:
        _KeysSynchronizer.automat(event, *args, **kwargs)
    return _KeysSynchronizer


#------------------------------------------------------------------------------


class KeysSynchronizer(automat.Automat):

    """
    This class implements all the functionality of ``keys_synchronizer()`` state machine.
    """

    fast = False

    def init(self):
        """
        Method to initialize additional variables and flags
        at creation phase of `keys_synchronizer()` machine.
        """

    def state_changed(self, oldstate, newstate, event, *args, **kwargs):
        """
        Method to catch the moment when `keys_synchronizer()` state were changed.
        """
        if newstate == 'NO_INFO' and event != 'instant':
            self.automat('instant')

    def A(self, event, *args, **kwargs):
        """
        The state machine code, generated using `visio2python <http://bitdust.io/visio2python/>`_ tool.
        """
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'init':
                self.state = 'NO_INFO'
                self.doInit(*args, **kwargs)
                self.SyncAgain = False
        #---NO_INFO---
        elif self.state == 'NO_INFO':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'run':
                self.state = 'RESTORE'
                self.doPrepare(*args, **kwargs)
                self.doRestoreKeys(*args, **kwargs)
            elif event == 'sync' or (event == 'instant' and self.SyncAgain):
                self.SyncAgain = False
                self.doSaveCallback(*args, **kwargs)
                self.doCheckAndRun(*args, **kwargs)
        #---IN_SYNC!---
        elif self.state == 'IN_SYNC!':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'disconnected':
                self.state = 'NO_INFO'
            elif event == 'sync' or (event == 'instant' and self.SyncAgain):
                self.state = 'RESTORE'
                self.SyncAgain = False
                self.doSaveCallback(*args, **kwargs)
                self.doPrepare(*args, **kwargs)
                self.doRestoreKeys(*args, **kwargs)
        #---RESTORE---
        elif self.state == 'RESTORE':
            if event == 'restore-ok':
                self.state = 'BACKUP'
                self.doBackupKeys(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'error' or event == 'disconnected':
                self.state = 'NO_INFO'
                self.doReportNoInfo(*args, **kwargs)
            elif event == 'sync':
                self.doSaveCallback(*args, **kwargs)
                self.SyncAgain = True
        #---BACKUP---
        elif self.state == 'BACKUP':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'backup-ok':
                self.state = 'CLEAN'
                self.doCleanKeys(*args, **kwargs)
            elif event == 'error' or event == 'disconnected':
                self.state = 'NO_INFO'
                self.doReportNoInfo(*args, **kwargs)
            elif event == 'sync':
                self.doSaveCallback(*args, **kwargs)
                self.SyncAgain = True
        #---CLEAN---
        elif self.state == 'CLEAN':
            if event == 'clean-ok':
                self.state = 'IN_SYNC!'
                self.doReportSync(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'disconnected':
                self.state = 'NO_INFO'
                self.doReportNoInfo(*args, **kwargs)
            elif event == 'sync':
                self.doSaveCallback(*args, **kwargs)
                self.SyncAgain = True
        #---CLOSED---
        elif self.state == 'CLOSED':
            pass
        return None

    def doInit(self, *args, **kwargs):
        """
        Action method.
        """
        self.result_callbacks = []

    def doSaveCallback(self, *args, **kwargs):
        """
        Action method.
        """
        if args and args[0]:
            self.result_callbacks.append(args[0])

    def doCheckAndRun(self, *args, **kwargs):
        """
        Action method.
        """
        from bitdust.customer import list_files_orator
        from bitdust.customer import fire_hire
        from bitdust.storage import backup_monitor
        list_files_orator_is_ready = list_files_orator.A().state == 'SAW_FILES'
        backup_monitor_is_ready = backup_monitor.A().state == 'READY'
        fire_hire_is_ready = fire_hire.A().state == 'READY'
        if list_files_orator_is_ready and backup_monitor_is_ready and fire_hire_is_ready:
            self.automat('run')
        else:
            reactor.callLater(5, self.automat, 'sync')  # @UndefinedVariable

    def doPrepare(self, *args, **kwargs):
        """
        Action method.
        """
        self.restored_count = 0
        self.saved_count = 0
        self.deleted_count = 0
        self.stored_keys = {}
        self.not_stored_keys = {}
        self.unreliable_keys = {}
        self.keys_to_upload = set()
        self.keys_to_erase = {}
        self.keys_to_rename = {}
        lookup = backup_fs.ListChildsByPath(path='.keys', recursive=False, backup_info_callback=restore_monitor.GetBackupStatusInfo)
        if isinstance(lookup, list):
            minimum_reliable_percent = eccmap.GetCorrectablePercent(eccmap.Current().suppliers_number)
            for i in lookup:
                if i['path'].endswith('.public'):
                    stored_key_id = i['path'].replace('.public', '').replace('.keys/', '')
                    is_private = False
                else:
                    stored_key_id = i['path'].replace('.private', '').replace('.keys/', '')
                    is_private = True
                if not my_keys.is_valid_key_id(stored_key_id):
                    lg.warn('not able to recognize stored key_id from item: %r' % i)
                    continue
                stored_key_id = my_keys.latest_key_id(stored_key_id)
                is_reliable = False
                for v in i['versions']:
                    try:
                        reliable = float(v['reliable'].replace('%', ''))
                    except:
                        lg.exc()
                        reliable = 0.0
                    if reliable >= minimum_reliable_percent:
                        is_reliable = True
                        break
                if is_reliable:
                    self.stored_keys[stored_key_id] = is_private
                else:
                    if is_private and my_keys.is_key_private(stored_key_id):
                        self.not_stored_keys[stored_key_id] = is_private
                    elif not is_private and my_keys.is_key_registered(stored_key_id):
                        self.not_stored_keys[stored_key_id] = is_private
                    else:
                        self.unreliable_keys[stored_key_id] = is_private
                    if _Debug:
                        lg.args(_DebugLevel, i=i)
        if _Debug:
            lg.args(_DebugLevel, stored_keys=len(self.stored_keys), not_stored_keys=list(self.not_stored_keys.keys()), unreliable_keys=len(self.unreliable_keys))

    def doRestoreKeys(self, *args, **kwargs):
        """
        Action method.
        """
        is_any_private_key_unreliable = bool(True in self.unreliable_keys.values())
        if is_any_private_key_unreliable and not self.stored_keys:
            if _Debug:
                lg.args(_DebugLevel, unreliable_keys=self.unreliable_keys)
            lg.err('not possible to restore any keys, all backup copies unreliable stored_keys=%d not_stored_keys=%d unreliable_keys=%d' % (len(self.stored_keys), len(self.not_stored_keys), len(self.unreliable_keys)))
            self.automat('error', Exception('not possible to restore any keys, all backup copies unreliable'))
            return
        keys_to_be_restored = []
        for key_id, is_private in self.stored_keys.items():
            latest_key_id = my_keys.latest_key_id(key_id)
            if latest_key_id != key_id:
                self.keys_to_rename[key_id] = (
                    latest_key_id,
                    is_private,
                )
            if my_keys.is_key_registered(key_id):
                if _Debug:
                    lg.out(_DebugLevel, '        skip restoring already known key_id=%r' % key_id)
                continue
            if my_keys.is_key_registered(latest_key_id):
                if _Debug:
                    lg.out(_DebugLevel, '        skip restoring already known latest key_id=%r' % latest_key_id)
                continue
            keys_to_be_restored.append((
                key_id,
                is_private,
            ))

        if _Debug:
            lg.args(_DebugLevel, keys_to_be_restored=len(keys_to_be_restored))

        def _on_restored_one(res, pos, key_id):
            self.restored_count += 1
            _do_restore_one(pos + 1)
            return None

        def _on_failed_one(err, pos, key_id):
            lg.err('failed to restore key %r : %r' % (key_id, err))
            _do_restore_one(pos + 1)
            return None

        def _do_restore_one(pos):
            if pos >= len(keys_to_be_restored):
                self.automat('restore-ok', True)
                return
            key_id, is_private = keys_to_be_restored[pos]
            res = key_ring.do_restore_key(key_id, is_private, wait_result=True)
            res.addCallback(_on_restored_one, pos, key_id)
            res.addErrback(_on_failed_one, pos, key_id)

        _do_restore_one(0)

    def doBackupKeys(self, *args, **kwargs):
        """
        Action method.
        """
        for old_key_id in list(self.keys_to_rename.keys()):
            new_key_id, is_private = self.keys_to_rename[old_key_id]
            if old_key_id in self.stored_keys and new_key_id not in self.stored_keys:
                self.keys_to_upload.add(new_key_id)
            if new_key_id in self.stored_keys and old_key_id in self.stored_keys:
                self.keys_to_erase[old_key_id] = is_private
        for key_id in my_keys.known_keys().keys():
            if key_id not in self.stored_keys or key_id in self.not_stored_keys:
                self.keys_to_upload.add(key_id)
        keys_saved = []
        for key_id in self.keys_to_upload:
            res = key_ring.do_backup_key(key_id)
            keys_saved.append(res)
            self.saved_count += 1
        if _Debug:
            lg.args(_DebugLevel, keys_saved=len(keys_saved))
        wait_all_saved = DeferredList(keys_saved, fireOnOneErrback=False, consumeErrors=True)
        wait_all_saved.addCallback(lambda ok: self.automat('backup-ok', ok))
        wait_all_saved.addErrback(lambda err: self.automat('error', err))

    def doCleanKeys(self, *args, **kwargs):
        """
        Action method.
        """
        keys_deleted = []
        for key_id, is_private in self.stored_keys.items():
            if not my_keys.is_key_registered(key_id):
                self.keys_to_erase[key_id] = is_private
        for key_id, is_private in self.keys_to_erase.items():
            res = key_ring.do_delete_key(key_id, is_private)
            keys_deleted.append(res)
            self.deleted_count += 1
        if _Debug:
            lg.args(_DebugLevel, restored=self.restored_count, saved=self.saved_count, deleted=self.deleted_count)
        self.automat('clean-ok')

    def doReportSync(self, *args, **kwargs):
        """
        Action method.
        """
        for cb in self.result_callbacks:
            if isinstance(cb, Deferred):
                cb.callback(True)
            else:
                cb(True)
        self.result_callbacks = []

    def doReportNoInfo(self, *args, **kwargs):
        """
        Action method.
        """
        if args and args[0]:
            err = args[0]
        else:
            err = Exception('failed to synchronize my keys')
        for cb in self.result_callbacks:
            if isinstance(cb, Deferred):
                cb.errback(err)
            else:
                cb(err)
        err_msg = ''
        if isinstance(err, failure.Failure):
            err_msg = err.getErrorMessage()
        else:
            err_msg = str(err)
        events.send('my-keys-synchronize-failed', data=dict(error=err_msg))
        self.result_callbacks = []

    def doDestroyMe(self, *args, **kwargs):
        """
        Remove all references to the state machine object to destroy it.
        """
        self.result_callbacks = []
        self.restored_count = 0
        self.saved_count = 0
        self.deleted_count = 0
        self.stored_keys = None
        self.keys_to_upload = None
        self.keys_to_erase = None
        self.keys_to_rename = None
        self.destroy()
        global _KeysSynchronizer
        del _KeysSynchronizer
        _KeysSynchronizer = None



#!/usr/bin/python
# data_sender.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (data_sender.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: data_sender.

.. raw:: html

    <a href="https://bitdust.io/automats/data_sender/data_sender.png" target="_blank">
    <img src="https://bitdust.io/automats/data_sender/data_sender.png" style="max-width:100%;">
    </a>

A state machine to manage data sending process, acts very simple:
    1) when new local data is created it tries to send it to the correct supplier
    2) wait while ``p2p.io_throttle`` is doing some data transmission to remote suppliers
    3) calls ``p2p.backup_matrix.ScanBlocksToSend()`` to get a list of pieces needs to be send
    4) this machine is restarted every minute to check if some more data needs to be send
    5) also can be restarted at any time when it is needed

EVENTS:
    * :red:`block-acked`
    * :red:`block-failed`
    * :red:`init`
    * :red:`new-data`
    * :red:`restart`
    * :red:`scan-done`
    * :red:`shutdown`
    * :red:`timer-1min`
    * :red:`timer-1sec`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 14

#------------------------------------------------------------------------------

import os
import time

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.automats import automat

from bitdust.lib import misc

from bitdust.lib import packetid

from bitdust.system import bpio

from bitdust.contacts import contactsdb

from bitdust.userid import my_id
from bitdust.userid import id_url

from bitdust.main import settings

from bitdust.p2p import online_status

from bitdust.stream import io_throttle

from bitdust.customer import list_files_orator

#------------------------------------------------------------------------------

STAT_KEEP_LATEST_RESULTS_COUNT = 5

_DataSender = None
_ShutdownFlag = False

#------------------------------------------------------------------------------


def A(event=None, *args, **kwargs):
    """
    Access method to interact with the state machine.
    """
    global _DataSender
    if _DataSender is None:
        _DataSender = DataSender(
            name='data_sender',
            state='AT_STARTUP',
            debug_level=_DebugLevel,
            log_events=_Debug,
            log_transitions=_Debug,
        )
    if event is not None:
        _DataSender.automat(event, *args, **kwargs)
    return _DataSender


class DataSender(automat.Automat):
    """
    A class to manage process of sending data packets to remote suppliers.
    """
    timers = {
        'timer-1min': (60, ['READY']),
        'timer-1sec': (1.0, ['SENDING']),
    }

    def A(self, event, *args, **kwargs):
        #---READY---
        if self.state == 'READY':
            if event == 'new-data' or event == 'timer-1min':
                self.state = 'SCAN_BLOCKS'
                self.doScanAndQueue(*args, **kwargs)
            elif event == 'restart':
                self.state = 'SCAN_BLOCKS'
                self.doCleanUpSendingQueue(*args, **kwargs)
                self.doScanAndQueue(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
        #---SCAN_BLOCKS---
        elif self.state == 'SCAN_BLOCKS':
            if event == 'scan-done' and self.isQueueEmpty(*args, **kwargs):
                self.state = 'READY'
                self.doRemoveUnusedFiles(*args, **kwargs)
            elif event == 'scan-done' and not self.isQueueEmpty(*args, **kwargs):
                self.state = 'SENDING'
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doCleanUpSendingQueue(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'restart':
                self.doCleanUpSendingQueue(*args, **kwargs)
                self.doScanAndQueue(*args, **kwargs)
        #---SENDING---
        elif self.state == 'SENDING':
            if event == 'restart':
                self.state = 'SCAN_BLOCKS'
                self.doCleanUpSendingQueue(*args, **kwargs)
                self.doScanAndQueue(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doCleanUpSendingQueue(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif (event == 'timer-1sec' or event == 'block-acked' or event == 'block-failed' or event == 'new-data') and self.isQueueEmpty(*args, **kwargs):
                self.state = 'SCAN_BLOCKS'
                self.doScanAndQueue(*args, **kwargs)
        #---AT_STARTUP---
        elif self.state == 'AT_STARTUP':
            if event == 'init':
                self.state = 'READY'
                self.doInit(*args, **kwargs)
        #---CLOSED---
        elif self.state == 'CLOSED':
            pass
        return None

    def isQueueEmpty(self, *args, **kwargs):
        if not args or not args[0] or not isinstance(args[0], list) or not isinstance(args[0], tuple):
            is_empty = io_throttle.IsSendingQueueEmpty()
            if _Debug:
                lg.out(_DebugLevel*2, 'data_sender.isQueueEmpty is_empty=%s' % is_empty)
            return is_empty
        remoteID, _ = args[0]
        can_send_to = io_throttle.OkToSend(remoteID)
        if _Debug:
            lg.out(_DebugLevel*2, 'data_sender.isQueueEmpty can_send_to=%s remoteID=%r' % (can_send_to, remoteID))
        return can_send_to

    def doInit(self, *args, **kwargs):
        """
        Action method.
        """
        self.statistic = {}

    def doScanAndQueue(self, *args, **kwargs):
        """
        Action method.
        """
        global _ShutdownFlag
        if _ShutdownFlag:
            if _Debug:
                lg.out(_DebugLevel, 'data_sender.doScanAndQueue   _ShutdownFlag is True')
            self.automat('scan-done', 0)
            return
        from bitdust.storage import backup_matrix
        from bitdust.storage import backup_fs
        backup_matrix.ReadLocalFiles()
        progress = 0
        # if _Debug:
        #     lg.out(_DebugLevel, 'data_sender.doScanAndQueue    with %d known customers' % len(contactsdb.known_customers()))
        for customer_idurl in contactsdb.known_customers():
            # if customer_idurl != my_id.getIDURL():
            #     # TODO: check that later
            #     lg.warn('skip sending to another customer: %r' % customer_idurl)
            #     continue
            known_suppliers = contactsdb.suppliers(customer_idurl)
            if not known_suppliers or id_url.is_some_empty(known_suppliers):
                if _Debug:
                    lg.out(_DebugLevel, 'data_sender.doScanAndQueue    found empty supplier(s) for customer %r, SKIP' % customer_idurl)
                continue
            known_backups = misc.sorted_backup_ids(list(backup_matrix.local_files().keys()), True)
            if _Debug:
                lg.out(_DebugLevel, 'data_sender.doScanAndQueue    found %d known suppliers for customer %r with %d backups' % (len(known_suppliers), customer_idurl, len(known_backups)))
            for backupID in known_backups:
                this_customer_idurl = packetid.CustomerIDURL(backupID)
                if this_customer_idurl != customer_idurl:
                    continue
                customerGlobalID, pathID, _ = packetid.SplitBackupID(backupID, normalize_key_alias=True)
                keyAlias = packetid.KeyAlias(customerGlobalID)
                item = backup_fs.GetByID(pathID, iterID=backup_fs.fsID(customer_idurl, keyAlias))
                if not item:
                    if _Debug:
                        lg.out(_DebugLevel, 'data_sender.doScanAndQueue    skip sending backup %r path not exist in catalog' % backupID)
                    continue
                if item.key_id and customerGlobalID and customerGlobalID != item.key_id:
                    if _Debug:
                        lg.out(_DebugLevel, 'data_sender.doScanAndQueue    skip sending backup %r key is different in the catalog: %r ~ %r' % (backupID, customerGlobalID, item.key_id))
                    continue
                packetsBySupplier = backup_matrix.ScanBlocksToSend(backupID, limit_per_supplier=None)
                total_for_customer = sum([len(v) for v in packetsBySupplier.values()])
                if total_for_customer:
                    if _Debug:
                        lg.out(_DebugLevel, 'data_sender.doScanAndQueue    sending %r for customer %r with %d pieces' % (item.name(), customer_idurl, total_for_customer))
                    for supplierNum in packetsBySupplier.keys():
                        # supplier_idurl = contactsdb.supplier(supplierNum, customer_idurl=customer_idurl)
                        if supplierNum >= 0 and supplierNum < len(known_suppliers):
                            supplier_idurl = known_suppliers[supplierNum]
                        else:
                            supplier_idurl = None
                        if not supplier_idurl:
                            lg.warn('skip sending, unknown supplier_idurl supplierNum=%s for %s, customer_idurl=%r' % (supplierNum, backupID, customer_idurl))
                            continue
                        for packetID in packetsBySupplier[supplierNum]:
                            backupID_, _, supplierNum_, _ = packetid.BidBnSnDp(packetID)
                            if backupID_ != backupID:
                                lg.warn('skip sending, unexpected backupID supplierNum=%s for %s, customer_idurl=%r' % (packetID, backupID, customer_idurl))
                                continue
                            if supplierNum_ != supplierNum:
                                lg.warn('skip sending, unexpected supplierNum %s for %s, customer_idurl=%r' % (packetID, backupID, customer_idurl))
                                continue
                            if io_throttle.HasPacketInSendQueue(supplier_idurl, packetID):
                                if _Debug:
                                    lg.out(_DebugLevel, 'data_sender.doScanAndQueue    %s already in sending queue for %r' % (packetID, supplier_idurl))
                                continue
                            latest_progress = self.statistic.get(supplier_idurl, {}).get('latest', '')
                            if len(latest_progress) >= 3 and latest_progress.endswith('---'):
                                if _Debug:
                                    lg.out(_DebugLevel + 2, 'data_sender.doScanAndQueue     skip sending to supplier %r because multiple packets already failed' % supplier_idurl)
                                continue
                            if not io_throttle.OkToSend(supplier_idurl):
                                if _Debug:
                                    lg.out(_DebugLevel + 2, 'data_sender.doScanAndQueue     skip sending, queue is busy for %r' % supplier_idurl)
                                continue
                            customerGlobalID, pathID = packetid.SplitPacketID(packetID)
                            filename = os.path.join(
                                settings.getLocalBackupsDir(),
                                customerGlobalID,
                                pathID,
                            )
                            if not os.path.isfile(filename):
                                if _Debug:
                                    lg.out(_DebugLevel, 'data_sender.doScanAndQueue     %s is not a file' % filename)
                                continue
                            itemInfo = item.to_json()
                            if io_throttle.QueueSendFile(
                                filename,
                                packetID,
                                supplier_idurl,
                                my_id.getIDURL(),
                                lambda packet, ownerID, packetID: self._packetAcked(packet, ownerID, packetID, itemInfo),
                                lambda remoteID, packetID, why: self._packetFailed(remoteID, packetID, why, itemInfo),
                            ):
                                progress += 1
                                if _Debug:
                                    lg.out(_DebugLevel, 'data_sender.doScanAndQueue   for %r put %s in the queue  progress=%d' % (item.name(), packetID, progress))
                            else:
                                if _Debug:
                                    lg.out(_DebugLevel, 'data_sender.doScanAndQueue    io_throttle.QueueSendFile FAILED %s' % packetID)
        if _Debug:
            lg.out(_DebugLevel, 'data_sender.doScanAndQueue    progress=%s' % progress)
        self.automat('scan-done', progress)


#     def doPrintStats(self, *args, **kwargs):
#         """
#         """
#        if lg.is_debug(18):
#            transfers = transport_control.current_transfers()
#            bytes_stats = transport_control.current_bytes_transferred()
#            s = ''
#            for info in transfers:
#                s += '%s ' % (diskspace.MakeStringFromBytes(bytes_stats[info.transfer_id]).replace(' ', '').replace('bytes', 'b'))
#            lg.out(0, 'transfers: ' + s[:120])

    def doRemoveUnusedFiles(self, *args, **kwargs):
        """
        Action method.
        """
        if not list_files_orator.is_synchronized():
            # always make sure we have a very fresh info about remote files before take any actions
            return
        # we want to remove files for this block
        # because we only need them during rebuilding
        if settings.getBackupsKeepLocalCopies() is True:
            # if user set this in settings - he want to keep the local files
            return
        # ... user do not want to keep local backups
        if settings.getGeneralWaitSuppliers() is True:
            from bitdust.customer import fire_hire
            # but he want to be sure - all suppliers are green for a long time
            if len(online_status.listOfflineSuppliers()) > 0 or (time.time() - fire_hire.GetLastFireTime() < 24*60*60):
                # some people are not there or we do not have stable team yet
                # do not remove the files because we need it to rebuild
                return
        count = 0
        from bitdust.storage import backup_matrix
        from bitdust.storage import restore_monitor
        from bitdust.storage import backup_rebuilder
        if _Debug:
            lg.out(_DebugLevel, 'data_sender.doRemoveUnusedFiles')
        for backupID in misc.sorted_backup_ids(list(backup_matrix.local_files().keys())):
            if restore_monitor.IsWorking(backupID):
                if _Debug:
                    lg.out(_DebugLevel, '        %s : SKIP, because restoring' % backupID)
                continue
            if backup_rebuilder.IsBackupNeedsWork(backupID):
                if _Debug:
                    lg.out(_DebugLevel, '        %s : SKIP, because needs rebuilding' % backupID)
                continue
            if not backup_rebuilder.ReadStoppedFlag():
                if backup_rebuilder.A().currentBackupID is not None:
                    if backup_rebuilder.A().currentBackupID == backupID:
                        if _Debug:
                            lg.out(_DebugLevel, '        %s : SKIP, because rebuilding is in process' % backupID)
                        continue
            if backupID not in backup_matrix.remote_files():
                if _Debug:
                    lg.out(_DebugLevel, '        going to erase %s because not found in remote files' % backupID)
                customer, pathID, version = packetid.SplitBackupID(backupID)
                dirpath = os.path.join(settings.getLocalBackupsDir(), customer, pathID, version)
                if os.path.isdir(dirpath):
                    try:
                        count += bpio.rmdir_recursive(dirpath, ignore_errors=True)
                    except:
                        lg.exc()
                continue
            packets = backup_matrix.ScanBlocksToRemove(backupID, check_all_suppliers=settings.getGeneralWaitSuppliers())
            for packetID in packets:
                customer, pathID = packetid.SplitPacketID(packetID)
                filename = os.path.join(settings.getLocalBackupsDir(), customer, pathID)
                if os.path.isfile(filename):
                    try:
                        os.remove(filename)
                    except:
                        lg.exc()
                        continue
                    count += 1
        if _Debug:
            lg.out(_DebugLevel, '    %d files were removed' % count)
        backup_matrix.ReadLocalFiles()

    def doCleanUpSendingQueue(self, *args, **kwargs):
        """
        Action method.
        """
        self.statistic = {}
        # io_throttle.DeleteAllSuppliers()

    def doDestroyMe(self, *args, **kwargs):
        """
        Action method.
        """
        self.statistic = {}
        self.destroy()
        global _DataSender
        del _DataSender
        _DataSender = None

    def _packetAcked(self, packet, ownerID, packetID, itemInfo):
        if _Debug:
            lg.args(_DebugLevel, pid=packetID, i=itemInfo)
        from bitdust.storage import backup_matrix
        backupID, blockNum, supplierNum, dataORparity = packetid.BidBnSnDp(packetID)
        backup_matrix.RemoteFileReport(backupID, blockNum, supplierNum, dataORparity, True, itemInfo)
        if ownerID not in self.statistic:
            self.statistic[ownerID] = {
                'acked': 0,
                'failed': 0,
                'latest': '',
            }
        self.statistic[ownerID]['acked'] += 1
        self.statistic[ownerID]['latest'] += '+'
        self.statistic[ownerID]['latest'] = self.statistic[ownerID]['latest'][-STAT_KEEP_LATEST_RESULTS_COUNT:]
        self.automat('block-acked', (ownerID, packetID))

    def _packetFailed(self, remoteID, packetID, why, itemInfo):
        if _Debug:
            lg.args(_DebugLevel, pid=packetID, i=itemInfo)
        from bitdust.storage import backup_matrix
        backupID, blockNum, supplierNum, dataORparity = packetid.BidBnSnDp(packetID)
        backup_matrix.RemoteFileReport(backupID, blockNum, supplierNum, dataORparity, False, itemInfo)
        if remoteID not in self.statistic:
            self.statistic[remoteID] = {
                'acked': 0,
                'failed': 0,
                'latest': '',
            }
        self.statistic[remoteID]['failed'] += 1
        self.statistic[remoteID]['latest'] += '-'
        self.statistic[remoteID]['latest'] = self.statistic[remoteID]['latest'][-STAT_KEEP_LATEST_RESULTS_COUNT:]
        self.automat('block-failed', (remoteID, packetID))


def statistic():
    """
    The ``data_sender()`` keeps track of sending results with every supplier.

    This is used by ``fire_hire()`` to decide how reliable is given
    supplier.
    """
    global _DataSender
    if _DataSender is None:
        return {}
    return _DataSender.statistic


def SetShutdownFlag(state=True):
    """
    Set flag to indicate that no need to send anything right now because service is stopping.
    """
    global _ShutdownFlag
    _ShutdownFlag = state



#!/usr/bin/python
# local_fs.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (local_fs.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: local_fs.


"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from io import open

#------------------------------------------------------------------------------

import os
import platform

#------------------------------------------------------------------------------

from bitdust.lib import strng

from bitdust.logs import lg

#------------------------------------------------------------------------------

_PlatformInfo = None

#------------------------------------------------------------------------------


def WriteBinaryFile(filename, data):
    """
    A smart way to write data to binary file. Return True if success.
    This should be atomic operation - data is written to another temporary file and than renamed.
    """
    global _PlatformInfo
    if _PlatformInfo is None:
        _PlatformInfo = platform.uname()
    try:
        tmpfilename = filename + '.new'
        f = open(tmpfilename, 'wb')
        bin_data = strng.to_bin(data)
        f.write(bin_data)
        f.flush()
        # from http://docs.python.org/library/os.html on os.fsync
        os.fsync(f.fileno())
        f.close()
        # in Unix the rename will overwrite an existing file,
        # but in Windows it fails, so have to remove existing file first
        if _PlatformInfo[0] == 'Windows' and os.path.exists(filename):
            os.remove(filename)
        os.rename(tmpfilename, filename)
    except:
        lg.exc('file write failed: %r' % filename)
        try:
            # make sure file gets closed
            f.close()
        except:
            pass
        return False
    return True


def AppendBinaryFile(filename, data, mode='a'):
    """
    Same as WriteBinaryFile but do not erase previous data in the file.
    TODO: this is not atomic right now
    """
    try:
        f = open(filename, mode)
        if 'b' in mode:
            bin_data = strng.to_bin(data)
            f.write(bin_data)
        else:
            f.write(data)
        f.flush()
        os.fsync(f.fileno())
        f.close()
    except:
        lg.exc()
        try:
            # make sure file gets closed
            f.close()
        except:
            lg.exc()
        return False
    return True


def ReadBinaryFile(filename, decode_encoding=None):
    """
    A smart way to read binary file. Return empty string in case of:

    - path not exist
    - process got no read access to the file
    - some read error happens
    - file is really empty
    """
    if not filename:
        return b''
    if not os.path.isfile(filename):
        return b''
    if not os.access(filename, os.R_OK):
        return b''
    try:
        infile = open(filename, mode='rb')
        data = infile.read()
        if decode_encoding is not None:
            data = data.decode(decode_encoding)
        infile.close()
        return data
    except:
        lg.exc('file read failed: %r' % filename)
        return b''


#------------------------------------------------------------------------------


def WriteTextFile(filepath, data):
    """
    A smart way to write data into text file. Return True if success.
    This should be atomic operation - data is written to another temporary file and than renamed.
    """
    temp_path = filepath + '.tmp'
    if os.path.exists(temp_path):
        if not os.access(temp_path, os.W_OK):
            return False
    if os.path.exists(filepath):
        if not os.access(filepath, os.W_OK):
            return False
        try:
            os.remove(filepath)
        except:
            lg.exc()
            return False
    fout = open(temp_path, 'wt', encoding='utf-8')
    text_data = strng.to_text(data)
    fout.write(text_data)
    fout.flush()
    os.fsync(fout)
    fout.close()
    try:
        os.rename(temp_path, filepath)
    except:
        lg.exc('file write failed: %r' % filepath)
        return False
    return True


def ReadTextFile(filename):
    """
    Read text file and return its content.
    """
    if not filename:
        return u''
    global _PlatformInfo
    if _PlatformInfo is None:
        _PlatformInfo = platform.uname()
    if not os.path.isfile(filename):
        return u''
    if not os.access(filename, os.R_OK):
        return u''
    try:
        infile = open(filename, 'rt', encoding='utf-8')
        data = infile.read()
        infile.close()
        return strng.to_text(data)
    except:
        lg.exc('file read failed: %r' % filename)
    return u''


#------------------------------------------------------------------------------


def RoundupFile(filename, stepsize):
    """
    For some things we need to have files which are round sizes, for example
    some encryption needs files that are multiples of 8 bytes.

    This function rounds file up to the next multiple of step size.
    """
    try:
        size = os.path.getsize(filename)
    except:
        lg.exc()
        return 0
    mod = size % stepsize
    increase = 0
    if mod > 0:
        increase = stepsize - mod
        fil = open(filename, 'a')
        fil.write(u' '*increase)
        fil.flush()
        os.fsync(fil)
        fil.close()
    return increase



"""
Modularize the mining algo check
"""

import hashlib
from decimal import Decimal
from quantizer import quantize_ten

__version__ = '0.0.1'


def bin_convert(string):
    return ''.join(format(ord(x), '8b').replace(' ', '0') for x in string)


def check_block(block_height_new, miner_address, nonce, db_block_hash, diff0, received_timestamp, q_received_timestamp, q_db_timestamp_last, peer_ip='N/A', app_log=None):
    """
    Checks that the given block matches the mining algo.

    :param block_height_new:
    :param miner_address:
    :param nonce:
    :param db_block_hash:
    :param diff0:
    :param received_timestamp:
    :param q_received_timestamp:
    :param q_db_timestamp_last:
    :param peer_ip:
    :param app_log:
    :return:
    """
    mining_hash = bin_convert(hashlib.sha224((miner_address + nonce + db_block_hash).encode('utf-8')).hexdigest())
    diff_drop_time = Decimal(180)
    mining_condition = bin_convert(db_block_hash)[0:int(diff0)]
    # simplified comparison, no backwards mining
    if mining_condition in mining_hash:
        if app_log:
            app_log.info('Difficulty requirement satisfied for block {} from {}'.format(block_height_new, peer_ip))
        diff_save = diff0

    elif Decimal(received_timestamp) > q_db_timestamp_last + Decimal(diff_drop_time):
        # uses block timestamp, don't merge with diff() for security reasons
        time_difference = q_received_timestamp - q_db_timestamp_last
        diff_dropped = quantize_ten(diff0) - quantize_ten(time_difference/diff_drop_time)
        if diff_dropped < 50:
            diff_dropped = 50

        mining_condition = bin_convert(db_block_hash)[0:int(diff_dropped)]
        if mining_condition in mining_hash:  # simplified comparison, no backwards mining
            if app_log:
                app_log.info('Readjusted difficulty requirement satisfied for block {} from {}'.format(block_height_new, peer_ip))
            diff_save = diff0
            # lie about what diff was matched not to mess up the diff algo
        else:
            raise ValueError('Readjusted difficulty too low for block {} from {}, should be at least {}'.format(block_height_new, peer_ip, diff_dropped))
    else:
        raise ValueError('Difficulty too low for block {} from {}, should be at least {}'.format(block_height_new, peer_ip, diff0))
    return diff_save



#!/usr/bin/python
# service_shared_data.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (service_shared_data.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: service_shared_data
"""

from __future__ import absolute_import
from bitdust.services.local_service import LocalService


def create_service():
    return SharedDataService()


class SharedDataService(LocalService):

    service_name = 'service_shared_data'
    config_path = 'services/shared-data/enabled'

    def dependent_on(self):
        return [
            'service_my_data',
        ]

    def start(self):
        from bitdust.main import events
        from bitdust.transport import callback
        from bitdust.access import shared_access_coordinator
        callback.append_inbox_callback(self._on_inbox_packet_received)
        events.add_subscriber(shared_access_coordinator.on_supplier_modified, 'supplier-modified')
        events.add_subscriber(shared_access_coordinator.on_my_list_files_refreshed, 'my-list-files-refreshed')
        events.add_subscriber(shared_access_coordinator.on_key_registered, 'key-registered')
        events.add_subscriber(shared_access_coordinator.on_key_erased, 'key-erased')
        events.add_subscriber(shared_access_coordinator.on_share_connected, 'share-connected')
        events.add_subscriber(shared_access_coordinator.on_supplier_file_modified, 'supplier-file-modified')
        shared_access_coordinator.open_known_shares()
        return True

    def stop(self):
        from bitdust.main import events
        from bitdust.transport import callback
        from bitdust.access import shared_access_coordinator
        events.remove_subscriber(shared_access_coordinator.on_key_registered, 'key-registered')
        events.remove_subscriber(shared_access_coordinator.on_key_erased, 'key-erased')
        events.remove_subscriber(shared_access_coordinator.on_share_connected, 'share-connected')
        events.remove_subscriber(shared_access_coordinator.on_my_list_files_refreshed, 'my-list-files-refreshed')
        events.remove_subscriber(shared_access_coordinator.on_supplier_modified, 'supplier-modified')
        events.remove_subscriber(shared_access_coordinator.on_supplier_file_modified, 'supplier-file-modified')
        callback.remove_inbox_callback(self._on_inbox_packet_received)
        return True

    def _on_inbox_packet_received(self, newpacket, info, status, error_message):
        from bitdust.p2p import commands
        from bitdust.access import key_ring
        if newpacket.Command == commands.Files():
            return key_ring.on_files_received(newpacket, info)
        return False



"""
websocket - WebSocket client library for Python

Copyright (C) 2010 Hiroki Ohtani(liris)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor,
    Boston, MA  02110-1335  USA

"""
import logging

_logger = logging.getLogger('websocket')
try:
    from logging import NullHandler
except ImportError:

    class NullHandler(logging.Handler):
        def emit(self, record):
            pass


_logger.addHandler(NullHandler())

_traceEnabled = False

__all__ = ['enableTrace', 'dump', 'error', 'warning', 'debug', 'trace', 'isEnabledForError', 'isEnabledForDebug', 'isEnabledForTrace']


def enableTrace(traceable, handler=logging.StreamHandler()):
    """
    turn on/off the traceability.

    traceable: boolean value. if set True, traceability is enabled.
    """
    global _traceEnabled
    _traceEnabled = traceable
    if traceable:
        _logger.addHandler(handler)
        _logger.setLevel(logging.DEBUG)


def dump(title, message):
    if _traceEnabled:
        _logger.debug('--- ' + title + ' ---')
        _logger.debug(message)
        _logger.debug('-----------------------')


def error(msg):
    _logger.error(msg)


def warning(msg):
    _logger.warning(msg)


def debug(msg):
    _logger.debug(msg)


def trace(msg):
    if _traceEnabled:
        _logger.debug(msg)


def isEnabledForError():
    return _logger.isEnabledFor(logging.ERROR)


def isEnabledForDebug():
    return _logger.isEnabledFor(logging.DEBUG)


def isEnabledForTrace():
    return _traceEnabled



from six import PY2, b


def method_factory_factory(method):
    def factory_py2(regex):
        from zope.interface.advice import addClassAdvisor
        _f = {}

        def decorator(f):
            _f[f.__name__] = f
            return f

        def advisor(cls):
            def wrapped(f):
                def __init__(self, *args, **kwargs):
                    f(self, *args, **kwargs)
                    func = None
                    orig = None
                    for func_name in _f:
                        orig = _f[func_name]
                        func = getattr(self, func_name)
                    if orig and func and func.im_func == orig:
                        self.register(method, regex, func)

                return __init__

            cls.__init__ = wrapped(cls.__init__)
            return cls

        addClassAdvisor(advisor)
        return decorator

    def factory_py3(regex):
        def decorator(f):
            current_methods = getattr(f, '__txrestapi__', [])
            current_methods.append((
                method,
                regex,
            ))
            f.__txrestapi__ = current_methods
            return f

        return decorator

    factory = factory_py2 if PY2 else factory_py3
    return factory


ALL = method_factory_factory(b('ALL'))
GET = method_factory_factory(b('GET'))
POST = method_factory_factory(b('POST'))
PUT = method_factory_factory(b('PUT'))
DELETE = method_factory_factory(b('DELETE'))



#!/usr/bin/python
# raidmake.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (make.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#

#
# Input is a file and a RAID spec.
# The spec says for each datablock which parity blocks XOR it.
# As we read in each datablock we XOR all the parity blocks for it.
# Probably best to read in all the datablocks at once, and do all
# the parities one time.  If we did one datablock after another
# the parity blocks would need to be active all the time.
#
# If we XOR integers, we have a byte order issue probably, though
# maybe not since they all stay the same order, whatever that is.
#
# Some machines are 64-bit.  Would be fun to make use of that if
# we have it.
#
# 2^1 => 3       ^ is XOR operator
#
# Parity packets have to be a little longer so they can hold
# parities on signatures of datapackets.  So we can reconstruct
# those signatures.
#
#

#------------------------------------------------------------------------------

from __future__ import absolute_import
from io import open

#------------------------------------------------------------------------------

_Debug = False

#------------------------------------------------------------------------------

import os
import sys
import copy
import array

#------------------------------------------------------------------------------

if __name__ == '__main__':
    dirpath = os.path.dirname(os.path.abspath(sys.argv[0]))
    sys.path.insert(0, os.path.abspath(os.path.join(dirpath, '..')))
    sys.path.insert(0, os.path.abspath(os.path.join(dirpath, '..', '..')))

#------------------------------------------------------------------------------

import bitdust.logs.lg

import bitdust.raid.eccmap
import bitdust.raid.raidutils

#------------------------------------------------------------------------------


def RoundupFile(filename, stepsize):
    """
    For some things we need to have files which are round sizes, for example
    some encryption needs files that are multiples of 8 bytes.

    This function rounds file up to the next multiple of step size.
    """
    try:
        size = os.path.getsize(filename)
    except:
        return
    mod = size % stepsize
    increase = 0
    if mod > 0:
        increase = stepsize - mod
        f = open(filename, 'ab')
        f.write(b' '*increase)
        f.close()


def ReadBinaryFile(filename):
    if not os.path.isfile(filename):
        return b''
    if not os.access(filename, os.R_OK):
        return b''
    f = open(filename, 'rb')
    data = f.read()
    f.close()
    return data


def WriteFile(filename, data):
    if sys.version_info[0] == 3:
        binary_type = bytes
    else:
        binary_type = str
    s = data
    if not isinstance(s, binary_type):
        s = s.encode('utf-8')
    f = open(filename, 'wb')
    f.write(s)
    f.close()


#------------------------------------------------------------------------------


def ReadBinaryFileAsArray(filename):
    if not os.path.isfile(filename):
        return b''
    if not os.access(filename, os.R_OK):
        return b''

    with open(filename, mode='rb') as f:
        values = array.array('i', f.read())

    values.byteswap()
    return values


def do_in_memory(filename, eccmapname, version, blockNumber, targetDir, threshold_control=None):
    try:
        if _Debug:
            with open('/tmp/raid.log', 'a') as logfile:
                logfile.write(u'make filename=%s eccmapname=%s blockNumber=%s\n' % (repr(filename), eccmapname, blockNumber))
        INTSIZE = 4
        myeccmap = bitdust.raid.eccmap.eccmap(eccmapname)
        # any padding at end and block.Length fixes
        RoundupFile(filename, myeccmap.datasegments*INTSIZE)
        wholefile = ReadBinaryFileAsArray(filename)
        length = len(wholefile)
        length = length*4
        seglength = (length + myeccmap.datasegments - 1)/myeccmap.datasegments

        # dict of data segments
        sds = {}
        for seg_num, chunk in enumerate(bitdust.raid.raidutils.chunks(wholefile, int(seglength/4))):
            FileName = targetDir + '/' + str(blockNumber) + '-' + str(seg_num) + '-Data'
            with open(FileName, mode='wb') as f:
                chunk_to_write = copy.copy(chunk)
                chunk_to_write.byteswap()
                sds[seg_num] = iter(chunk)
                f.write(chunk_to_write)

        psds_list = bitdust.raid.raidutils.build_parity(
            sds,
            int(seglength/INTSIZE),
            myeccmap.datasegments,
            myeccmap,
            myeccmap.paritysegments,
            threshold_control=threshold_control,
        )

        dataNum = len(sds)
        parityNum = len(psds_list)

        for PSegNum, _ in psds_list.items():
            FileName = targetDir + '/' + str(blockNumber) + '-' + str(PSegNum) + '-Parity'
            with open(FileName, mode='wb') as f:
                f.write(psds_list[PSegNum])

        return dataNum, parityNum

    except:
        bitdust.logs.lg.exc()
        return -1, -1


def main():
    do_in_memory(filename=sys.argv[1], eccmapname=sys.argv[2], version=sys.argv[3], blockNumber=int(sys.argv[4]), targetDir=sys.argv[5])


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# authority_node.py
#
"""
.. module:: authority_node
.. role:: red

BitDust authority_node() Automat

EVENTS:
    * :red:`init`
    * :red:`no-record`
    * :red:`read-failed`
    * :red:`run-cycle`
    * :red:`shutdown`
    * :red:`timer-1min`
    * :red:`tx-failed`
    * :red:`tx-not-found`
    * :red:`tx-sent`
    * :red:`valid-record`
    * :red:`valid-tx-found`
"""

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 12

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.lib import strng

from bitdust.automats import automat

from bitdust.main import config

from bitdust.blockchain import bismuth_wallet

from bitdust.dht import dht_records

from bitdust.contacts import identitycache

from bitdust.userid import id_url

#------------------------------------------------------------------------------

_AuthorityNode = None

#------------------------------------------------------------------------------


def A(event=None, *args, **kwargs):
    """
    Access method to interact with `authority_node()` machine.
    """
    global _AuthorityNode
    if event is None:
        return _AuthorityNode
    if _AuthorityNode is None:
        if event == 'shutdown':
            return _AuthorityNode
        # TODO: set automat name and starting state here
        _AuthorityNode = AuthorityNode(
            name='authority_node',
            state='AT_STARTUP',
            debug_level=_DebugLevel,
            log_events=_Debug,
            log_transitions=_Debug,
            publish_events=False,
        )
    if event is not None:
        _AuthorityNode.automat(event, *args, **kwargs)
    return _AuthorityNode


#------------------------------------------------------------------------------


class AuthorityNode(automat.Automat):

    """
    This class implements all the functionality of ``authority_node()`` state machine.
    """

    timers = {
        'timer-1min': (60, ['IDLE']),
    }

    def __init__(self, debug_level=0, log_events=False, log_transitions=False, publish_events=False, **kwargs):
        """
        Builds `authority_node()` state machine.
        """
        self.dht_position = 0
        self.dht_last_value = None
        self.empty_streak = 0
        super(AuthorityNode, self).__init__(
            debug_level=debug_level,
            log_events=log_events,
            log_transitions=log_transitions,
            publish_events=publish_events,
            **kwargs,
        )

    def __repr__(self):
        return '%s[%d|%d](%s)' % (self.id, self.dht_position, self.empty_streak, self.state)

    def init(self):
        """
        Method to initialize additional variables and flags
        at creation phase of `authority_node()` machine.
        """

    def state_changed(self, oldstate, newstate, event, *args, **kwargs):
        """
        Method to catch the moment when `authority_node()` state were changed.
        """

    def state_not_changed(self, curstate, event, *args, **kwargs):
        """
        This method intended to catch the moment when some event was fired in the `authority_node()`
        but automat state was not changed.
        """

    def A(self, event, *args, **kwargs):
        """
        The state machine code, generated using `visio2python <https://github.com/vesellov/visio2python>`_ tool.
        """
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'init':
                self.state = 'IDLE'
                self.doInit(*args, **kwargs)
                self.doNextRecord(*args, **kwargs)
                self.doRunCycleSoon(*args, **kwargs)
        #---IDLE---
        elif self.state == 'IDLE':
            if event == 'run-cycle':
                self.state = 'DHT_READ'
                self.doDHTRead(*args, **kwargs)
            elif event == 'timer-1min':
                self.doRunCycleSoon(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
        #---DHT_READ---
        elif self.state == 'DHT_READ':
            if event == 'valid-record':
                self.state = 'FIND_TX'
                self.doFindTransaction(*args, **kwargs)
            elif event == 'no-record' or event == 'read-failed':
                self.state = 'IDLE'
                self.doNextRecord(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
        #---FIND_TX---
        elif self.state == 'FIND_TX':
            if event == 'tx-not-found':
                self.state = 'SEND_TX'
                self.doSendTransaction(*args, **kwargs)
            elif event == 'valid-tx-found':
                self.state = 'IDLE'
                self.doDHTErase(*args, **kwargs)
                self.doNextRecord(*args, **kwargs)
                self.doRunCycleSoon(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
        #---SEND_TX---
        elif self.state == 'SEND_TX':
            if event == 'tx-sent' or event == 'tx-failed':
                self.state = 'IDLE'
                self.doNextRecord(*args, **kwargs)
                self.doRunCycleSoon(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
        #---CLOSED---
        elif self.state == 'CLOSED':
            pass
        return None

    def doInit(self, *args, **kwargs):
        """
        Action method.
        """
        self.dht_position = config.conf().getInt('services/blockchain-authority/requests-reading-offset', 0)
        self.dht_last_value = None
        self.empty_streak = 0

    def doNextRecord(self, *args, **kwargs):
        """
        Action method.
        """
        reading_limit = config.conf().getInt('services/blockchain-authority/requests-reading-limit', 10)
        empty_streak_limit = min(5, int(reading_limit/5.0) + 1)
        if (self.empty_streak >= empty_streak_limit) or (self.dht_position > reading_limit):
            self.dht_position = config.conf().getInt('services/blockchain-authority/requests-reading-offset', 0)
            self.empty_streak = 0
        self.dht_last_value = None
        self.dht_position += 1

    def doRunCycleSoon(self, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.args(_DebugLevel, dht_position=self.dht_position, empty_streak=self.empty_streak)
        self.automat('run-cycle')

    def doDHTRead(self, *args, **kwargs):
        """
        Action method.
        """
        d = dht_records.get_bismuth_identity_request(position=self.dht_position, use_cache=False)
        d.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='authority_node.doDHTRead')
        d.addCallback(self._on_dht_read_result, self.dht_position)
        d.addErrback(self._on_dht_read_failed)

    def doDHTErase(self, *args, **kwargs):
        """
        Action method.
        """
        dht_records.erase_bismuth_identity_request(position=self.dht_position)

    def doFindTransaction(self, *args, **kwargs):
        """
        Action method.
        """
        clean_public_key = strng.to_text(self.dht_last_value['public_key'])[:10000].replace('ssh-rsa ', '')
        idurl = id_url.field(self.dht_last_value['idurl'])
        results = bismuth_wallet.find_transaction(
            recipient=self.dht_last_value['wallet_address'],
            operation='identity',
            openfield='{}:{}'.format(idurl.username, clean_public_key),
        )
        if results:
            self.automat('valid-tx-found')
        else:
            self.automat('tx-not-found')

    def doSendTransaction(self, *args, **kwargs):
        """
        Action method.
        """
        registration_bonus_coins = config.conf().getInt('services/blockchain-authority/registration-bonus-coins', 0)
        clean_public_key = strng.to_text(self.dht_last_value['public_key'])[:10000].replace('ssh-rsa ', '')
        idurl = id_url.field(self.dht_last_value['idurl'])
        try:
            tx_id = bismuth_wallet.send_transaction(
                recipient=self.dht_last_value['wallet_address'],
                amount=registration_bonus_coins,
                operation='identity',
                data='{}:{}'.format(idurl.username, clean_public_key),
            )
        except:
            lg.exc()
            self.automat('tx-failed')
            return
        self.automat('tx-sent', tx_id)

    def doDestroyMe(self, *args, **kwargs):
        """
        Remove all references to the state machine object to destroy it.
        """
        self.destroy()
        global _AuthorityNode
        del _AuthorityNode
        _AuthorityNode = None

    def _on_identity_cached(self, src, idurl, value):
        ident = identitycache.FromCache(idurl)
        if not id_url.is_the_same(ident.getIDURL(), value['idurl']):
            lg.warn('ivalid Bismuth identity register request, idurl is not matching')
            self.empty_streak += 1
            self.dht_last_value = None
            self.automat('read-failed', value)
            return
        if strng.to_text(ident.getPublicKey()).replace('ssh-rsa ', '') != value['public_key']:
            lg.warn('ivalid Bismuth identity register request, public key is not matching')
            self.empty_streak += 1
            self.dht_last_value = None
            self.automat('read-failed', value)
            return
        self.dht_last_value = value
        self.empty_streak = 0
        self.automat('valid-record', value)

    def _on_identity_cache_failed(self, err, idurl):
        if _Debug:
            lg.args(_DebugLevel, err=err)
        self.empty_streak += 1
        self.dht_last_value = None
        self.automat('read-failed', err)

    def _on_dht_read_result(self, value, position):
        if _Debug:
            lg.args(_DebugLevel, position=position, value=value)
        if not value:
            self.empty_streak += 1
            self.dht_last_value = value
            self.automat('no-record', value)
            return
        try:
            idurl = value['idurl']
            value['public_key']
            value['wallet_address']
        except Exception as exc:
            lg.exc()
            self.empty_streak += 1
            self.dht_last_value = value
            self.automat('read-failed', exc)
            return
        d = identitycache.immediatelyCaching(idurl)
        d.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='authority_node._on_dht_read_result')
        d.addErrback(self._on_identity_cache_failed, idurl)
        d.addCallback(self._on_identity_cached, idurl, value)

    def _on_dht_read_failed(self, result):
        if _Debug:
            lg.args(_DebugLevel, result=result)
        self.empty_streak += 1
        self.dht_last_value = result
        self.automat('no-record', result)
